#!/usr/bin/env python
# encoding: utf-8
import os
import sys
import time
import glob
import shutil

from multiprocessing import Pool
from subprocess import CalledProcessError
import traceback

import logging
LOG = logging.getLogger(__name__)
import pprint
pp = pprint.PrettyPrinter(indent=4)

from adl_common import ADL_HOME
from adl_common import ADL_VARS
from adl_common import env
from adl_common import sh
from adl_common import get_return_code

from cspp_some_utils.log_common import status_line 
from granule_status import Granule_processing_status
from system_checks import System_checks


from sdr_pre_process import Sdr_pre_process

from sdr_process import Sdr_process
from compress_and_aggregate import run_aggregation

from product_operations import Product_operations
from adl_log import checkADLLogForSuccess
from compress_and_aggregate import repack_products
from shortnames import Shortnames
from cspp_some_utils.basics import check_and_convert_env_var

from random import *

class ADL_SDR:
    """
    """
    work_dir=None
    product_output_dir=None

    allow_cache_update=None
    instrument=None
    spacecraft=None
    sdr_products=None
    agg_products=None
    geo_products=None
    controllers=[]    
    controllers_xml=[]
    fnxmls = [] 
    
    trim=[0,0]
    nproc=1
    
    center=False
    accuracyMode='reduced'
    cleanup=True
    
    apply_size_threshold = True
    expected_granule_delta = None
    GIG_MEMORY_PER_PROC = None
    
    granule_status = None
    shortname = None
    deploy=0
    run_cmo = False
    
    def __init__(cls,work_dir,allow_cache_update,instrument,spacecraft):
        """
        """
        cls.work_dir=work_dir
        cls.allow_cache_update=allow_cache_update
        cls.instrument=instrument
        cls.spacecraft=spacecraft
        LOG.info(cls.__dict__)
        cls.shortname= Shortnames()
        # for now pull out of environment
        if 'CSPP_SDR_OUTPUT_DIR' in os.environ:
            product_output_dir = check_and_convert_env_var('CSPP_SDR_OUTPUT_DIR', check_write=True,
                                                      default_value=None,create=False,silent=True)

            cls.product_output_dir=product_output_dir
            
        
    def cleanup_work_dir(cls,force_all=False):
        "upon successful run, clean out work directory"
        
        remaining_files_and_directories = glob.glob(os.path.join(cls.work_dir, '%s*' % cls.spacecraft))

        for r in remaining_files_and_directories:
            shutil.rmtree(r)
            
        # remove the granule end singnal directories
        remaining_files_and_directories = glob.glob(os.path.join(cls.work_dir, 'CSPP_SDR_GRAN*' ))

        for r in remaining_files_and_directories:
            shutil.rmtree(r)    
            
        LOG.debug('remaining_files_and_directories %d' % len(remaining_files_and_directories))
        if len(remaining_files_and_directories) != 0 or force_all is True:
            
            for d in "msd" , "log" , "dynamic"  ,"logs" , "db" : 
                dir_to_go = os.path.join(cls.work_dir, d)       
                if os.path.exists(dir_to_go):
                    shutil.rmtree(dir_to_go)
            
            dir_to_go = os.path.join(cls.work_dir, "*.log")
            if os.path.exists(dir_to_go):
                shutil.rmtree(dir_to_go)

            
        tlef = glob.glob(os.path.join(cls.work_dir,'TLE*.*'))
        for tle in tlef: 
            os.remove(tle)
            
        tlef = glob.glob(os.path.join(cls.work_dir,'*.msd'))
        for tle in tlef: 
            os.remove(tle)
        
        tin = glob.glob(os.path.join(cls.work_dir,'TLINSET'))
        for tle in tin: 
            os.remove(tle)
    
        tlef = glob.glob(os.path.join(cls.work_dir,'CrIS-SDR-ENGPKT-BACKUP-AUX*.*'))
        for tle in tlef:
            os.remove(tle)
        
        tlef = glob.glob(os.path.join(cls.work_dir,'CrIS-FS-SDR-ENGPKT-BACKUP*.*'))
        for tle in tlef:
            os.remove(tle)

        tlef = glob.glob(os.path.join(cls.work_dir,'*Matrix*.*'))
        for tle in tlef:
            os.remove(tle)       
        
        pixes = glob.glob(os.path.join(cls.work_dir,'*.pkl'))
        for pix in pixes:
            os.remove(pix)   
    
    def proc_cfg(cls,trim,nproc) :
        """
        """
        cls.trim = trim
        cls.nproc = int(nproc)
        

    def devel_cfg(cls,accuracyMode,cleanup,center,apply_size_threshold):
        """
        """
        cls.accuracyMode = accuracyMode
        cls.cleanup = cleanup
        cls.center = center
        cls.apply_size_threshold = apply_size_threshold
           
    def product_cfg(cls,aggregate,compress):
        """
        """  
        
        cls.aggregate = aggregate
        cls.compress= bool(compress) 
     
    def reconstruct(cls,granule_output_dir):
        """
        """

    def check_cmo(cls,argument_dictionaries,granules_to_process):
        """
        """
        return argument_dictionaries

    def run_and_check_single_granule(cls,additional_env):
        "run a VIIRS SDR XML input in sequence"


        work_dir = additional_env['WORK_DIR']
        xml = additional_env['XML_FILE']
        xml_files_str = additional_env['XML_FILE']
        xml_files = xml_files_str.split(":")
        granule_id = additional_env['N_Granule_ID']
        granule_output_dir = additional_env['ADL_HDF_OUTPUT_PATH']

        #compress = bool(additional_env['COMPRESS'])
        compress = bool(cls.compress)
         
        cleanup = additional_env['CLEANUP']  
        cleanup = cls.cleanup  

        chunk_size = int(additional_env['CHUNK_SIZE'])
        start_time = time.time()
        
        product_operations = Product_operations(granule_output_dir,
                                                cls.instrument,
                                                cls.spacecraft,
                                                product_output_dir=cls.product_output_dir)
        
        # if cls.instrument == 'VIIRS' :
        #     st=cls.deploy
        #     LOG.info("sleep %d"%int(st))
        #     if st != 0 :
        #         time.sleep(st)
        #     cls.deploy+=1
                
        good = True
        try:
            for controller in range(0,len(cls.controllers)) :

                cmd = [cls.controllers[controller], xml_files[controller]]
                status_line('Executing %r with WORK_DIR=%r ' % (cmd, work_dir))
                LOG.debug('additional environment vars: %s' % 
                            repr(additional_env))   
                z = ADL_VARS.copy()
                z.update(additional_env)  # which returns None since it mutates z
                pid = sh(cmd, env=env(**z), cwd=work_dir)
                LOG.debug("%r ran as pid %d" % (cmd, pid))
        # if not check_log_files(work_dir, pid, xml):
        #            bad_log_runs.add(granule_id)

        except CalledProcessError as oops:
            pid = getattr(oops, 'pid', None)
            #        if not check_log_files(work_dir, pid, xml):
            #            bad_log_runs.add(granule_id)

            LOG.debug(traceback.format_exc())
            LOG.error('%s failed on %r: %r. Continuing...' % (cmd[0], xml, oops))
            #        crashed_runs.add(granule_id)
            good = False

        product_shortnames = sorted( cls.agg_products + cls.geo_products )
        product_ids = []

        for short_name in product_shortnames:
            sdr_name = cls.shortname.SHORTNAME_2_PRODUCTID[short_name]
            product_ids.append(sdr_name)


        problem = product_operations.check_products(granule_id, 
                                                    product_shortnames,
                                                    chunk_size=chunk_size)

        if problem is True:
            good = False
            
        if not checkADLLogForSuccess(granule_output_dir, xml_files[0]):
            good = False    
        
        #product_operations.rename_products( cls.sdr_products)
        product_operations.rename_products( product_ids)

        gran_id = additional_env['N_Granule_ID']

        product_operations.add_geo_attribute_to_h5(
             gran_id, cls.agg_products)
        
        cls.reconstruct(granule_output_dir)

        if compress is True:
            status_line('Compress products for %s' % granule_id)
            repack_products(granule_output_dir, product_shortnames)

        chunks , option_chunks = product_operations.move_products_to_work_directory( 
                                                                product_ids)

        end_time = time.time()
        elapsed_time = end_time - start_time
        status_line('Tot Execution Time:  "%f" Gran "%s"' %
                    (elapsed_time, granule_id))

        stat = ""
        if good is True:
            stat = 'Granule %s Completed without problems. ' % granule_id

        if good is False:
            stat = 'Granule %s Completed with problems. ' % granule_id

        cls.granule_status.report_gran_completion(stat,chunks,int(additional_env["ACTUAL"]))

        result = {"GOOD": str(good)}

        return result


    def run_all_granules(cls,
                        granules_to_process, 
                        no_chunk=True,                     
                        **additional_env):
        """run each OMPS SDR XML input in sequence
        return the list of granule IDs which crashed, and list of granule ids which did not create output
        """


        noncritical_problem = 0
        start_time = time.time()

        # has products in it! Starting with a clean working directory is strongly recommended.")
        sdr_proc = Sdr_process(cls.work_dir,cls.instrument,cls.spacecraft)
        total_granules = len(granules_to_process)
    
        inproc= int( cls.nproc )
        fudge = inproc
        if int(  total_granules % inproc ) == 0 :
            fudge = 0
        chunk_size= int(  (total_granules + fudge )  / inproc )
        if chunk_size <= 0:
            chunk_size=1
        if cls.instrument == 'CRIS':
            no_chunk = False            
            
        if no_chunk is True:    
            chunk_size=1    
            
        additional_keys = dict(spacecraft=cls.spacecraft,
                            CHUNK_SIZE=str(chunk_size),
                            INSTRUMENT=cls.instrument)

        argument_dictionaries, granules_to_process = sdr_proc.setup_multi_process(
                                                                        granules_to_process,
                                                                        cls.fnxmls,
                                                                        cls.controllers_xml,
                                                                        additional_keys,
                                                                        compress=cls.compress,
                                                                        cleanup=cls.cleanup,
                                                                        accuracyMode=cls.accuracyMode)
        
        
   

        cls.granule_status = Granule_processing_status(total_granules)


        argument_dictionaries = cls.check_cmo(argument_dictionaries,
                      granules_to_process)


        results = []
        
        end_time = time.time()
        elapsed_time = end_time - start_time
        status_line('Setup Execution Time:  "%f" ' % (elapsed_time))
    
        environment_error = False
        if len(argument_dictionaries) > 0:

            sys_checks = System_checks(int(cls.nproc),cls.GIG_MEMORY_PER_PROC)
            nprocs = sys_checks.number_processors_used

            parallel = Pool(int(nprocs))

            try:
                # results = parallel.map_async(run_and_check_single_granule,
                #                              argument_dictionaries).get(9999999)
                results = parallel.map(cls.run_and_check_single_granule,
                                            argument_dictionaries)
                
            except KeyboardInterrupt:
                LOG.error("Got exception, stopping workers and exiting.")
                parallel.terminate()
                parallel.join()
                sys.exit(1)

        noncritical_problem += sdr_proc.finish_and_cleanup_multiprocessing(results,argument_dictionaries,cls.cleanup)    
        return noncritical_problem

    def pre_process( cls, h5_names ):
        """
        """
        
        sdr_pp = Sdr_pre_process(cls.work_dir,
                                 cls.instrument,
                                 cls.spacecraft)
        
        unpacking_problems, granules_to_process, cls.trim = sdr_pp.pre_process_rdr(
            h5_names,cls.trim,
            apply_size_threshold=cls.apply_size_threshold,
            center=cls.center)

        LOG.debug(', '.join(x['N_Granule_ID'] for x in granules_to_process))
        if not granules_to_process:
            LOG.warning("Found no granules to process!")
            LOG.warning("inputs %s"%str(h5_names))
            return get_return_code(unpacking_problems, 
                                   num_xml_files_to_process=0, 
                                   num_no_output_runs=0,
                                    noncritical_problem=False, 
                                    environment_error=0)


        ereturn = sdr_pp.required_ancillary(granules_to_process,
                                  allow_cache_update=cls.allow_cache_update)
        
        if ereturn == None:
            cls.run_cmo = True
        total_granules = len(granules_to_process)
        #LOG.info('%d granules to process' % (len(granules_to_process)))

        granules_to_process = sdr_pp.trim_the_granules(
            granules_to_process, cls.expected_granule_delta, trim=cls.trim)
        
        return granules_to_process, unpacking_problems


    def sdr( cls, 
             h5_names):
        
        "process OMPS RDR data in HDF5 format (any aggregation) to SDRs in an arbitrary work directory"
  
        if not cls.trim:
            cls.trim = [0, 0]
            
        granules_to_process, unpacking_problems = cls.pre_process( h5_names  )    

        try:
            
            # build XML configuration files for jobs that can be run

            LOG.debug("running SDR on XML files")
            
            noncritical_problem = cls.run_all_granules( 
                                                       granules_to_process, 
                                                        ADL_HOME=ADL_HOME)
            

            product_shortnames = cls.agg_products + cls.geo_products


            if cls.aggregate is True:
                number_problems = run_aggregation(cls.work_dir,cls.nproc,cls.spacecraft,
                                cls.agg_products,cls.geo_products) 
                
                # product output dir cannot be used with aggregation
                product_operations = Product_operations(cls.work_dir,
                                                cls.instrument,
                                                cls.spacecraft,
                                                product_output_dir=None)
                
                product_ids = []

                for short_name in product_shortnames:
                    sdr_name = cls.shortname.SHORTNAME_2_PRODUCTID[short_name]
                    product_ids.append(sdr_name)
            
                product_operations.rename_products( product_ids)
        
                # number_problems += product_operations.add_geo_attribute_to_aggregates(
                #         product_shortnames)
                
                product_operations.add_geo_attribute_to_h5( "Agg", product_shortnames)
        
                if number_problems > 0:
                    LOG.info('Number aggregation problems %d' % int(number_problems))
                
                

            # print final disposition message and get return code
            total_granules = len(granules_to_process)
            gran_failure = total_granules - cls.granule_status.number_good_granules()
            rc = get_return_code(unpacking_problems, 
                                 total_granules, 
                                 gran_failure, 
                                 noncritical_problem,
                                environment_error=False)

        except EnvironmentError:
            LOG.error(traceback.format_exc())
            # print final disposition message and get return code
            rc = get_return_code(unpacking_problems,
                                 len(granules_to_process),
                                 num_no_output_runs=0,
                                noncritical_problem=False,
                                environment_error=True)

        return rc
 
        
        